﻿/////////////////////////////////////////////////////////////////////////////
// Copyright 2016 Charlie Wartnaby    charlie@wartnaby.org
//
//   This is open source. The formal licence here covers things like
//   me not accepting any liability for problems, but it is a permissive
//   licence, i.e. allowing this code to be reused in your own programs
//   even if they are commercial.
//
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
/////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using System.Xml;

namespace GarminGpxHrmToCsv
{
    public partial class MainForm : Form
    {
        internal struct Datapoint
        {
            internal DateTime Time;
            internal double HeartRateBpm;

            internal Datapoint(DateTime dt, double hr)
            {
                Time = dt;
                HeartRateBpm = hr;
            }
        };

        private string InputFilePath = "";
        private XmlDocument GpxFileContent = null;
        private List<Datapoint> PointList = null;
        private Thread ProcessingThread = null;
        private List<string> WaitingMessages = new List<string>();
        private bool ProcessingInProgress = false;


        public MainForm()
        {
            InitializeComponent();
            UpdateButtons();
        }

        private void buttonOpenGpxOrAbort_Click(object sender, EventArgs e)
        {
            if (ProcessingInProgress)
            {
                // Abort
                if (null != ProcessingThread) ProcessingThread.Abort();
                Thread.Sleep(100); // let it finish tidying up
            }
            else
            {
                // Select file and start
                using (OpenFileDialog ofd = new OpenFileDialog())
                {
                    ofd.Filter = "GPX/TCX files (*.gpx;*.tcx)|*.gpx;*.tcx";
                    ofd.Multiselect = false;

                    if (ofd.ShowDialog() == DialogResult.OK)
                    {
                        // User wants to proceed, so kick off processing
                        InputFilePath = ofd.FileName;

                        richTextBoxStatus.Clear();

                        if (ProcessingThread != null)
                        {
                            // Kill existing processing if in progress already (shouldn't be anyhow)
                            ProcessingThread.Abort();
                            Thread.Sleep(100);
                        }

                        ProcessingThread = new Thread(FileLoadProcessBackgroundWorker);
                        ProcessingThread.IsBackground = true;
                        ProcessingThread.Start();
                        ProcessingInProgress = true;
                        UpdateButtons();
                    }
                    else
                    {
                        // User cancelled
                    }
                }
            }
        }

        private void FileLoadProcessBackgroundWorker()
        {
            try
            {
                GpxFileContent = new XmlDocument();
                GpxFileContent.Load(InputFilePath);
            }
            catch (Exception e)
            {
                AddMessage("ERROR: GPX/TCX load failed: " + e.Message + Environment.NewLine);
                return;
            }

            AddMessage("Attempting to parse data..." + Environment.NewLine);

            bool is_tcx = InputFilePath.ToLower().EndsWith("tcx");

            // Processing of the order of 10000 points is quite slow, so be prepared to be
            // interrupted by the user aborting
            try
            {
                ProcessDatapoints(is_tcx);
            }
            catch (Exception ex)
            {
                if (ex is ThreadAbortException || ex is ThreadInterruptedException)
                {
                    AddMessage(Environment.NewLine + "ERROR: processing aborted" + Environment.NewLine);
                }
                else
                {
                    AddMessage(Environment.NewLine + "ERROR: unexpected problem: " + ex.Message + Environment.NewLine);
                }
                PointList.Clear();
            }
            finally
            {
                ProcessingInProgress = false;
            }
        }

        private void ProcessDatapoints(bool is_tcx)
        {
            // Use points from any segments within track
            string trackpoint_tag = is_tcx ? "Trackpoint" : "trkpt";

            XmlNodeList trackPoints = GpxFileContent.DocumentElement.SelectNodes("//*[local-name()='" + trackpoint_tag + "']");

            if (trackPoints.Count < 1)
            {
                AddMessage("ERROR: cannot find any '" + trackpoint_tag + "' entries in file");
                return;
            }

            AddMessage(trackPoints.Count + " track points found, processing..." + Environment.NewLine);

            PointList = new List<Datapoint>();
            int errorCount = 0;
            bool progressDotsDrawn = false;

            foreach (XmlNode trackPointNode in trackPoints)
            {
                XmlElement elPoint = trackPointNode as XmlElement;
                if (elPoint == null) continue; // shouldn't happen

                XmlElement elTime, elHR;

                if (is_tcx)
                {
                    elTime = elPoint.SelectSingleNode("*[local-name()='Time']") as XmlElement;
                    XmlElement elHrParent = elPoint.SelectSingleNode(".//*[local-name()='HeartRateBpm']") as XmlElement;
                    if (elHrParent == null) continue;
                    elHR = elHrParent.SelectSingleNode("./*[local-name()='Value']") as XmlElement;
                }
                else
                {
                    elTime = elPoint.SelectSingleNode("*[local-name()='time']") as XmlElement;
                    elHR = elPoint.SelectSingleNode(".//*[local-name()='hr']") as XmlElement;
                }

                if (elTime == null || elHR == null) continue;

                string timeStr = elTime.InnerText;
                string hrStr = elHR.InnerText;

                DateTime time;
                double hr;

                try
                {
                    time = DateTime.Parse(timeStr);
                    hr = double.Parse(hrStr);
                    PointList.Add(new Datapoint(time, hr));
                }
                catch
                {
                    errorCount++;
                }

                if (PointList.Count % 200 == 0)
                {
                    // Show progress in case we're taking a while so we don't just look frozen
                    AddMessage(".");
                    progressDotsDrawn = true;
                }
            }

            if (progressDotsDrawn) AddMessage(Environment.NewLine);

            if (PointList.Count < 1)
            {
                AddMessage("ERROR: Failed to parse any time/date values in track points" + Environment.NewLine);
                return;
            }

            if (errorCount > 0)
            {
                AddMessage("WARNING: failed to extract time/date values from " + errorCount + " track points" + Environment.NewLine);
            }

            AddMessage("Read " + PointList.Count + " heart rate points OK" + Environment.NewLine);
        }

        // callable from background thread
        private void AddMessage(string text)
        {
            lock (WaitingMessages)
            {
                WaitingMessages.Add(text);
            }
        }

        private void UpdateButtons()
        {
            buttonOpenGpxOrAbort.Text = (ProcessingInProgress ? "Ab&ort" : "&Open and Process GPX File...");
            buttonSaveAs.Enabled = !ProcessingInProgress && (PointList != null) && (PointList.Count > 0);
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            // Can only update GUI from main thread
            lock (WaitingMessages)
            {
                while (WaitingMessages.Count > 0)
                {
                    richTextBoxStatus.AppendText(WaitingMessages[0]);
                    WaitingMessages.RemoveAt(0);
                }
            }

            // If background processing finishes, will need to update which
            // controls can be used and their text
            UpdateButtons();
        }

        private void buttonSaveAs_Click(object sender, EventArgs e)
        {
            using (SaveFileDialog sfd = new SaveFileDialog())
            {
                sfd.FileName = Path.GetFileNameWithoutExtension(InputFilePath);
                sfd.Filter = "Comma-Separated Values (*.csv)|*.csv";

                if (sfd.ShowDialog() == DialogResult.OK)
                {
                    try
                    {
                        StreamWriter sw = new StreamWriter(sfd.FileName);
                        sw.WriteLine("Time of day,Time since start [sec],Heart rate [bpm]");

                        bool first = true;
                        DateTime firstTime = DateTime.Now;

                        foreach (Datapoint dp in PointList)
                        {
                            if (first)
                            {
                                firstTime = dp.Time;
                                first = false;
                            }

                            TimeSpan relativeTime = dp.Time - firstTime;

                            sw.WriteLine(dp.Time.ToLongTimeString() + "," + relativeTime.TotalSeconds.ToString() + "," + dp.HeartRateBpm);
                        }

                        sw.Close();
                        MessageBox.Show("File written OK");
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show("ERROR: " + ex.Message);
                    }
                }
            }
        }

        private void buttonExit_Click(object sender, EventArgs e)
        {
            this.Close();
        }
    }
}
